/* Cache.pmod
 *
 * De olika typer av cachar som anvnds av databasen. */

#define WERROR werror

class Large
{
  inherit Thread.Mutex;
  mapping(int:mixed) cache;

  mixed lookup(int key)
  {
    return cache[key];
  }

  void remember(int key, mixed value)
  {
    cache[key] = value;
  }

  void forget(int key)
  {
    m_delete(cache, key);
  }

  array all_indices()
  {
    return indices(cache);
  }

  array all_values()
  {
    return values(cache);
  }
  
  void create()
  {
    cache = ([ ]);
  }  
}

class Small
{
  inherit Thread.Mutex;
  array(mixed) cache;

  mixed lookup(int key)
  {
    int h = key % sizeof(cache);
    array v = cache[h];
    return v && (v[0] == key) && v[1];
  }

  void remember(int key, mixed value)
  {
    int h = key % sizeof(cache);
    cache[h] = ({ key, value });
  }

  void forget(int key)
  {
    int h = key % sizeof(cache);
    array v = cache[h];
    if (v && (v[0] == key))
      cache[h] = 0;
  }
  
  void create(int size)
  {
    cache = allocate(size || 4711);
  }
}

class Node
{
  // object(Node) prev;
  // object(Node) head;
  object prev;
  object next;
  mixed value;

  void create(mixed o)
  {
    value = o;
  }
}

#define CHECK(x) check_state(x)

class Queue
{
  /* Sorted queue */
  object(Node) head;
  object(Node) tail;

  int size;

  void check_state(object | void node)
  {
    if (!head)
      {
	if (tail)
	  throw( ({ "Cache.Queus->check_state: head = 0, tail != 0!\n",
		      backtrace() }) );
	if (size)
	  throw( ({ "Cache.Queus->check_state: head = tail = 0, size != 0!\n",
		      backtrace() }) );
	if (node)
	  throw( ({ "Cache.Queus->check_state: "
		      "Empty queue, node not found!\n",
		      backtrace() }) );
	return;
      }
    int found = 0;
    int s = 1;
    object p = head;
    if (p->prev)
      throw( ({ "Cache.Queus->check_state: "
		  "Head node has non-zero prev pointer!\n",
		  backtrace() }) );
	
    while(1)
      {
	if (p == node)
	  found = 1;
	if (object_program(p) != Node)
	  throw( ({ "Cache.Queus->check_state: none-Node in queue!\n",
		      backtrace() }) );
	if (!p->next)
	  {
	    if (tail != p)
	      throw( ({ "Cache.Queus->check_state: "
			  "Tail node is not pointed to by tail!\n",
			  backtrace() }) );
	    break;
	  }
	if (p != p->next->prev)
	  throw( ({ "Cache.Queus->check_state: Invalid prev pointer!\n",
			  backtrace() }) );
	p = p->next;
	s++;
	if (s > (size + 17))
	  throw( ({ sprintf("Cache.Queus->check_state: "
			    "Queue way larger than %d !\n", size) }) );
      }
    if (s != size)
      throw( ({ sprintf("Cache.Queus->check_state: "
			"Expected size: %d, actual: %d.\n",
			size, s) , backtrace() }) );
    if (node && !found)
	  throw( ({ "Cache.Queus->check_state: "
		      "Node not found!\n",
		      backtrace() }) );
    
    return;
  }
	    
  object add_head(mixed o)
  {
    CHECK(0);
    object(Node) n = Node(o);
    if (head)
      head->prev = n;
    n->next = head;
    head = n;
    if (!tail)
      tail = n;
    size++;
    CHECK(n);
    return n;
  }

  object remove_tail()
  {
    CHECK(0);
    if (!tail)
      return 0;
    object o = tail;
    tail = o->prev;
    if (o->prev)
      o->prev->next = 0;
    else
      head = 0;
    size--;
    CHECK(0);
    return o;
  }

  void remove(object(Node) node)
  {
    CHECK(node);
    if (node->prev)
      node->prev->next = node->next;
    else
      head = node->next;
    if (node->next)
      node->next->prev = node->prev;
    else
      tail = node->prev;
    size--;
    CHECK(0);
  }
  
  void move_to_head(object n)
  {
    CHECK(n);
    if (head == n)
      return;

    /* Ta bort fr}n listan */
    n->prev->next = n->next;
    if (n->next)
      n->next->prev = n->prev;
    else
      tail = n->prev;

    /* Lnka in frst */
    head->prev = n;
    n->next = head;
    
    n->prev = 0;
    head = n;
    CHECK(0);
  }

  void create()
  {
    head = tail = 0;
    size = 0;
    CHECK(0);
  }
}

class Dirt
{
  inherit Thread.Condition : dirty;
  inherit Queue : queue;
  
  object(Node) mark_dirty(mixed value)
  {
    dirty::signal();
    return queue::add_head(value);
  }

  void mark_clean(object(Node) n)
  {
    queue::remove(n);
  }

  mixed get_dirty()
  {
    object(Node) o = queue::remove_tail();
    return o && o->value;
  }

  void wait_dirty(object key)
  {
    dirty::wait(key);
  }
  
  void create()
  {
    queue::create();
  }
}

class LRU
{
  inherit Thread.Mutex;
  inherit Thread.Condition : full;

  /* Innehller noder ({ index, vrde, dirt-node }) */
  inherit Queue : lru;

  /* Innehller queue-noder */
  inherit Dirt : dirt;

  /* Innehller samma noder som lru:: */
  mapping(int:object) cache;
  
  int max_size;
  int inertia;
  int inertia_count;

  object(Node) lookup_node(int key, int reorder)
  {
    object n = cache[key];
    if (n && reorder)
      {
	if (inertia == inertia_count)
	  {
	    lru::move_to_head(n);
	    inertia_count = 0;
	  }
	else
	  inertia_count++;
      }
    return n;
  }
  
  mixed lookup(int key)
  {
    object n = lookup_node(key, 1);
    return n && n->value[1];
  }

  int is_full()
  {
    return lru::size >= max_size;
  }

  void remember(int key, mixed value)
  {
    if (is_full())
      full::signal(); /* Kan vcka en trd som kan garba
		       * cachen. */
    object n = lru::add_head( ({ key, value, 0 }) );
    cache[key] = n;
  }

  void update(int key, mixed value)
  {
    object n = lookup_node(key, 0);
    if (!n)
      throw( ({ "Cache.LRU->update: No such key.\n", backtrace() }) );
    if (n->value[2])
      throw( ({ "Cache.LRU->update: Update on dirty node not expected!\n",
		  backtrace() }) );
    n->value[1] = value;
  }

  void update_dirty(int key, mixed value)
  {
    object n = lookup_node(key, 0);
    if (!n)
      throw( ({ "Cache.LRU->update: No such key.\n", backtrace() }) );
    n->value[1] = value;
    if (!n->value[2])
      { /* Behver sparas, pga denna ndring. */
	n->value[2] = dirt::mark_dirty(n);
      }
  }
	
  void forget(int key)
  {
    object n = cache[key];
    if (n)
      {
	if (n->value[2])
	  dirt::mark_clean(n->value[2]);
	m_delete(cache, key);
	lru::remove(n);
      }
  }

  void wait_full(object key)
  {
    full::wait(key);
  }
  
  array garb_item()
  {
    object o = lru::remove_tail();
    if (o)
      {
	m_delete(cache, o->value[0]);
	if (o->value[2])
	  dirt::mark_clean(o->value[2]);
      }
    return o && o->value;
  }

  array sync_item()
  {
    object(Node) n = dirt::get_dirty();
    if (n)
      {
	n->value[2] = 0;
	return n->value[..1];
      }
    return 0;
  }
  
  void create(int size)
  {
    max_size = size;
    lru::create();
    dirt::create();
    cache = ([ ]);
  }
}

    
class Random
{
  // inherit Large;
  inherit Thread.Mutex;
  inherit Thread.Condition : full;
  inherit Dirt : dirt;

  mapping cache;
  int max_size;

  int is_full()
  {
    return sizeof(cache) >= max_size;
  }

  mixed lookup(int key)
  {
    array a = cache[key];
    return a && a[1];
  }

  void remember(int key, mixed value)
  {
    if (is_full())
      full::signal();
    cache[key] = ({ key, value, 0 });
  }

  void update(int key, mixed value)
  {
    array a = cache[key];
    if (!a)
      throw( ({ "Cache.Random->update: No such key.\n", backtrace() }) );
    if (a[2])
      throw( ({ "Cache.LRU->update: Update on dirty node not expected!\n",
		  backtrace() }) );
    a[1] = value;
  }

  void update_dirty(int key, mixed value)
  {
    array a = cache[key];
    if (!a)
      throw( ({ "Cache.Random->update: No such key.\n", backtrace() }) );
    a[1] = value;
    if (!a[2])
      { /* Behver sparas, pga denna ndring. */
	a[2] = dirt::mark_dirty(a);
      }
  }
	
  void forget(int key)
  {
    array a = cache[key];
    if (a)
      {
	if (a[2])
	  dirt::mark_clean(a[2]);

	m_delete(cache, key);
      }
  }

  void wait_full(object key)
  {
    full::wait(key);
  }

  array garb_item()
  {
    int size = sizeof(cache);
    if (!size)
      return 0;
    int i = random(size);
    int key = indices(cache)[i];
    array a = cache[key];
    if (a[2])
      dirt::mark_clean(a[2]);
    m_delete(cache, key);
    return a;  /* ({ key, value, dirty }) */
  }

  array sync_item()
  {
    array a = dirt::get_dirty();
    if (a)
      {
	a[2] = 0;
	return a[..1];
      }
    return 0;
  }
  
  void create(int size)
  {
    max_size = size;
    cache = ([ ]);
  }
}
  
class GC_etc
{
  object cache;
  function(int, mixed:void) flush;
  float delay;
  
  object gc_tid;
  object sync_tid;
  
  void run_gc()
  {
    object key = cache->lock();
    while(1)
      {
	if (cache->is_full())
	  {
	    WERROR("Garbage collecting one item...\n");
	    array a = cache->garb_item();
	    if (a && flush && a[2])
	      {
		WERROR("Collected item synced first.\n");
		flush(@a[..1]);
	      }
	    key = 0;      /* Lt andra trdar kra */
	    key = cache->lock();
	  }
	else
	  cache->wait_full(key);
      }
  }

  void run_sync()
  {
    object key = cache->lock();
    while(1)
      {
	WERROR("Sync running.\n");
	array a = cache->sync_item();
	if (a)
	  {
	    WERROR("Syncing one item...\n");
	    flush(@a);
	    key = 0;  /* Lt andra trdar kra */
	    WERROR("Sync sleeping.\n");
	    sleep(delay);
	    key = cache->lock();
	  }
	else
	  {
	    WERROR("Sync waiting.\n");
	    cache->wait_dirty(key);
	  }
      }
  }
  
  void create(object c, function(int, mixed:void)|void f,
	      float|int|void t)
  {
    cache = c;
    flush = f;
    gc_tid = thread_create(run_gc);
    if (flush)
      {
	delay = (float) t / cache->max_size;
	sync_tid = thread_create(run_sync);
      }
  }
}
