/* Database.pmod
 *
 */

#define WERROR werror

/* Lsordning
 *
 * Fr att undvika deadlocks, s r reglerna:
 *
 * Undvik att lsa mer n ett ls i taget. Om du mste, s r
 * ordningen: Man fr lsa sessionsobjekt efter att ha lst
 * sessions-tabellen, men inte tvrtom.
 *
 * Efter att ha lst ett sessionsobjekt, s fr man lsa ett ls till,
 * s lnge det inte r sessionstabellen eller ett annat
 * sessionsobjekt.
 *
 * Samma gller membershipstabellen. Man fr lsa den, och sedan lsa
 * mten och personer, men inte tvrtom. */

/* Tabeller: */

  
object texts;
object text_stats;
object uconferences;
object conferences;
object conf_texts;
object upersons;
object memberships;
object created_texts;
object unread;
object marks;
object globals;
object sessions;

object session_counter;

object db;

class Table
{
  object table;
  object cache;
  program type;
  
  string make_key(int n)
  {
    return sprintf("%4c", n);
  }

  array(int) db_indices()
  {
    array(string) a = indices(table);
    // WERROR(sprintf("Database.Table->db_indices: %O\n", a));
    return Array.map(a, lambda(string s) {
      int i;
      if ((1 != sscanf(s, "%4c", i)) || (i <= 0))
	{
	  werror("Database.Table->indices: Bad index!\n");
	  return 0;
	}
      return i;
    }) - ({ 0 });
  }

  array(int) cache_indices()
  {
    return cache->all_indices();
  }

  array cache_values()
  {
    return cache->all_values();
  }
  
  mixed unlocked_lookup(int n)
  {
    mixed v = cache->lookup(n);
    if (v)
      return v;
    v = table->get(make_key(n));
    if (v)
      {
	mixed o;
	if (!type)
	  o = v;
	else
	  {
	    if (!arrayp(v) || (sizeof(v) != 2))
	      {
		werror("Database.Table->unlocked_lookup: "
		       "Bad type in database!\n");
		return 0;
	      }
	    o = type();
	    if (o->db_version == v[0])
	      o = o->from_db(v[1]);
	    else
	      {
		werror("Database.Table: "
		       "Attempting to version promote data.\n");
		if (!o->from_db_version)
		  {
		    werror("Database.Table->unlocked_lookup: "
			   "Version mismatch.\n");
		    return 0;
		  }
		o = o->from_db_version(v[0], v[1]);
	      }
	    if (!o)
	      {
		werror("Database.Table->unlocked_lookup: "
		       "Invalid entry in database!\n");
		return 0;
	      }
	  }
	cache->remember(n, o);
	return o;
      }
    return 0;
  }

  void delete(int n)
  {
    object key = cache->lock();
    cache->forget(n);
    table->delete(make_key(n));
  }
  
  /* Cachen r lst s lnge det returnerade vrdet finns kvar. */
  array lookup(int n)
  {
    object key = cache->lock();
    object v = unlocked_lookup(n);
    if (v)
      return ({ v, key });
    else return 0;
  }

  array query_vars(int n, array(string) vars)
  {
    object key = cache->lock();
    object v = unlocked_lookup(n);
    return v && rows(v, vars);
  }
  
  /* Ska anropas med cachen lst.
   * Sparar vrdet p disk, utan att anvnda cachen.
   * Anropas typiskt frn gc- eller sync-trden. */
  void update_db(int n, mixed v)
  {
    // WERROR(sprintf("Database.Table: update_db(%d, %O)\n", n, v));
    table->set(make_key(n), type ? ({ v->db_version, v->to_db() }) : v);
    // WERROR(sprintf("indices = %O\n", indices(table)));
  }

  /* Ska anropas med cachen lst */
  void update(int n, mixed v)
  {
    cache->remember(n, v);
    update_db(n, v);
  }

  /* Ska anropas med cachen lst.
   * Sparar inte ut det nya vrdet p disk. */
  void update_cache(int n, mixed v)
  {
    cache->update_dirty(n, v);
  }

  void unlocked_store(int n, mixed v)
  {
    update(n, v);
  }
  
  /* Store a new value */
  void store(int n, mixed v)
  {
    object key = cache->lock();
    unlocked_store(n, v);
  }

  void store_new(int n)
  {
    store(n, type());
  }
  
  void create(object t, object c, program|void p)
  {
    table = t;
    cache = c;
    type = p;
  }
}

class Counter
{
  inherit Thread.Mutex;
  int n;

  int current()
  {
    object key = lock();
    return n;
  }
    
  int next()
  {
    object key = lock();
    n++;
    return n;
  }

  void create(int start)
  {
    n = start;
  }
}

class Text_db
{
  inherit Table;

  void create(object db, int cache_size)
  {
    Table::create(db->table("texts"), Cache.Small(cache_size),
		  Server.Text);
  }

  string query_text(int n)
  {
    array a = lookup(n);
    return a && a[0]->text;
  }
}

class Text_Stat_db
{
  inherit Table;

  int query_author(int n)
  {
    array a = lookup(n);
    return a ? a[0]->author : -1;
  }

  string query_subject(int n)
  {
    array a = lookup(n);
    return a && a[0]->subject;
  }
  
  array(int) query_all_recipients(int n, int person)
  {
    array a = lookup(n);
    if (!a)
      return 0;

    return column(a[0]->recipients + a[0]->cc_recipients, 0);
  }

  int add_comment_in(int n, array comment)
  {
    array a = lookup(n);
    if (!a)
      return 0;

    a[0]->commented_ins += ({ comment });
    update(n, a[0]);
    return 1;
  }
  
  int add_footnote_in(int n, array footnote)
  {
    array a = lookup(n);
    if (!a)
      return 0;

    a[0]->footnoted_ins += ({ footnote });
    update(n, a[0]);
    return 1;
  }

  void create(object db, int cache_size)
  {
    Table::create(db->table("text_stats"), Cache.Small(cache_size),
		  Server.Text_Stat);
  }
}

class UConf_db
{
  inherit Table;
  // array(int) existing_confs;
  
  void create(object db)
  {
    Table::create(db->table("uconfs"), Cache.Large(),
		  Server.UConference);
    array(int) confs = db_indices();
    WERROR(sprintf("Database.UConf_db->create: confs = %O\n", confs));
    /* Fyll cachen */
    foreach(confs, int conf)
      lookup(conf);
  }

  int query_type(int n)
  {
    array a = lookup(n);
    return a ? a[0]->type : -1;
  }
  
  int is_secret(int n)
  {
    array a = lookup(n);
    return a && (a[0]->type & Types.conf_secret);
  }

  /* r mtet hemligt eller lskyddat? */
  int is_protected(int n)
  {
    array a = lookup(n);
    return a && (a[0]->type & (Types.conf_secret | Types.conf_rd_prot) );
  }
  array query_canonical_names()
  {
    object key = cache->lock();
    return ({ mkmultiset(Array.map(column(cache_values(), "name"),
				   Misc.kom_canonicalize_name)), key });
  }

  array(object) query_z_info()
  {
    object key = cache->lock();
    return Array.map(cache_indices(), lambda(int n)
		     {
		       /* Man skulle kunna anvnda samma klass fr
			* uconf och conf-z-info, men man vill nd
			* kopiera datat fr att kunna slppa
			* lset p uconf-cachen. */
		       object info = Server.Conf_Z_Info();
		       object uconf = unlocked_lookup(n);
		       if (!uconf)
			 {
			   werror("Database.UConf_db->query_z_info: "
				  "lookup failed!\n");
			   return 0;
			 }
		       info->name = uconf->name;
		       info->type = uconf->type;
		       info->conf_no = n;
		       return info;
		     } ) - ({ 0 });
  }

  int create_conf(int n, object(Server.UConference) conf)
  {
    // WERROR(sprintf("Database.UConf_db->create_conf: %O\n",
    // mkmapping(indices(conf), values(conf))));
	   
    array a = query_canonical_names();
    WERROR(sprintf("exixting names: %O\n", a[0]));
    if (a[0][Misc.kom_canonicalize_name(conf->name)])
      /* Namnet finns redan */
      return 0;
    unlocked_store(n, conf);
    WERROR(sprintf("create_conf: cache_indices = %O,\n"
		   "             db_indices = %O\n",
		   cache_indices(), db_indices()));
    return 1;
  }
}

class Conf_db
{
  inherit Table;

  object gc;
  
  void create(object db)
  {
    Table::create(db->table("confs"), Cache.Large(),
		  Server.Conference);
  }

  int query_supervisor(int n)
  {
    array a = lookup(n);
    return a ? a[0]->supervisor : -1;
  }

  int is_member(int n, int person)
  {
    array a = lookup(n);
    return a && a[0]->members[person];
  }

  int add_member(int n, int person)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    a[0]->members |= (< person >);
    update(n, a[0]);
    return 1;
  }
  
  int may_write_to_conf(int n, int person)
  {
    array a = lookup(n);
    if (!a)
      return 0;

    int permitted_submitters = a[0]->permitted_submitters;
    if (!permitted_submitters)
      return 1;
    a = 0;

    return is_member(permitted_submitters, person);
  }
}

class Conf_texts_db
{
  inherit Table;

  object gc;
  
  void create(object db, int cache_size)
  {
    Table::create(db->table("conf_texts"), Cache.Random(cache_size),
		  Server.Conf_texts);
    gc = Cache.GC_etc(cache);
  }
  
  int query_first_text(int n)
  {
    array a = lookup(n);
    return a && sizeof(a[0]->texts) && a[0]->texts[0];
  }
    
  int query_last_text(int n)
  {
    array a = lookup(n);

    return a && sizeof(a[0]->texts) && a[0]->texts[-1];
  }

  int query_no_texts(int n)
  {
    array a = lookup(n);

    return a && sizeof(a[0]->texts);
  }
    
  array(int) query_texts_after(int n, int text_no)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    int i = Misc.binary_search(text_no, a[0]->texts);
    return a[0]->texts[i+1..];
  }

  array(int) query_texts_interval(int n, int start, int end)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    if (!sizeof(a[0]->texts) || (start > a[0]->texts[-1]) )
      return ({ });
    int i = Misc.binary_search(start - 1, a[0]->texts) + 1;
    int j = Misc.binary_search(end, a[0]->texts);
    WERROR(sprintf("query_texts_interval: [i..j] = [%d..%d]\n",
		   i, j));
    if (j<i)
      /* Inga texter i intervallet. */
      return ({ });
    return a[0]->texts[i..j];
  }
  
  int add_text(int n, int text_no)
  {
    array a = lookup(n);
    if (!a)
      return 0;
  
    /* Det kan ha skapats fler texter sedan textnumret allokerades,
     * s det r inte skert att texten kan lggas till p slutet
     * av den sorterade listan. */

    a[0]->texts = Misc.sorted_insert(a[0]->texts, text_no);
    update(n, a[0]);

    return sizeof(a[0]->texts);
  }  
}

/* Data om personer som aldrig kastas bort nr de vl lsts in,
 * ungefr persontypen i protokoll A. */
class UPerson_db
{
  inherit Table;

  void create(object db)
  {
    Table::create(db->table("upersons"), Cache.Large(),
		  Server.UPerson);
  }

  string query_passwd(int n)
  {
    array a = lookup(n);
    return a && a[0]->passwd;
  }

  int exists(int n)
  {
    return !!lookup(n);
  }
  
  int record_login(int n, string username)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    a[0]->username = username;
    a[0]->last_login = time();
    a[0]->sessions++;
    update(n, a[0]);
    return 1;
  }

  int record_created_conf(int n)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    a[0]->created_confs++;
    update(n, a[0]);
    return 1;
  }
       
  int record_created_pers(int n)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    a[0]->created_persons++;
    update(n, a[0]);
    return 1;
  }

  int record_created_text(int n)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    a[0]->no_of_created_texts++;
    update(n, a[0]);
    return 1;
  }

  int record_memberships(int n, int no_of_confs)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    a[0]->no_of_confs = no_of_confs;
    update(n, a[0]);
    return 1;
  }
}

/* Information som r mer skrymmande. */
class Memberships_db
{
  inherit Table;

  object gc;

  void create(object db, int cache_size, float|int sync_interval)
  {
    Table::create(db->table("memberships"), Cache.Random(cache_size),
		  Server.Person_memberships);
    gc = Cache.GC_etc(cache, update_db, sync_interval);
  }

  array query_membership(int n, int conf)
  {
    array a = lookup(n);
    if (!a)
      return 0;

    foreach(a[0]->memberships, object m)
      {
	if (m->conference == conf)
	  return ({ m, @a });
      }
    return 0;
  }

  /* ka first_unread endast om ingen annan har hunnit fre */
  int increase_first_unread(int n, int conf, int change_count, int value)
  {
    array a = query_membership(n, conf);
    if (!a)
      return 0;
    if (a[0]->change_count != change_count)
      return 1;
    a[0]->change_count++;
    a[0]->first_unread = value;
    update_cache(n, a[1]);
    return 1;
  }

  /* Minska first_unread */
  int decrease_first_unread(int n, int conf, int value)
  {
    array a = query_membership(n, conf);
    if (!a)
      return 0;
    if (a[0]->first_unread > value)
      {
	a[0]->first_unread = value;
	a[0]->change_count++;
	update(n, a[1]);
      }
  }

}

/* Skapade texter */
class Person_texts_db
{
  inherit Table;

  object gc;

  int add_text(int n, int text_no)
  {
    array a = lookup(n);
    if (!a)
      return 0;
    /* Det kan ha skapats fler texter sedan textnumret allokerades,
     * s det r inte skert att texten kan lggas till p slutet
     * av den sorterade listan. */
    a[0]->texts = Misc.sorted_insert(a[0]->texts, text_no);
    update(n, a[0]);
    return 1;
  }

  void create(object db, int cache_size)
  {
    Table::create(db->table("person_texts"), Cache.Random(cache_size),
		  Server.Person_texts);
    gc = Cache.GC_etc(cache);
  }
}

/* Information som mest r intressant nr personen r inloggad, som
 * den potentiellt *stora* olsta listan. */
class Person_unread_db
{
  inherit Table;

  object gc;
  
  void create(object db, int cache_size, float|int sync_interval)
  {
    Table::create(db->table("person_unread"), Cache.LRU(cache_size),
		  Server.Person_unread);
    gc = Cache.GC_etc(cache, update_db, sync_interval);
  }

  int mark_as_read(int person, array(int) texts)
  {
    array a = lookup(person);
    if (!a)
      return Errors.undef_pers;

    int changed = 0;
    foreach(texts, int text)
      {
	if (a[0]->mark_as_read(text))
	  changed = 1;
      }
    if (changed)
      update_cache(person, a[0]);
    return 0;
  }
}

class Mark_db
{
  inherit Table;

  object gc;

  void create(object db, int cache_size)
  {
    Table::create(db->table("marks"), Cache.Random(cache_size),
		  Server.Person_marks);
    gc = Cache.GC_etc(cache);
  }
}

class Session_db
{
  inherit Cache.Large : cache;

  array lookup(int n)
  {
    WERROR("Session_db->lookup\n");
    object key = lock();
    object v = cache::lookup(n);
    object key2 = v->lock();
    return v && ({ v->session, key, key2 });
  }

#if 0
  /* Returnerar -1 om sessionen inte fanns */
  int query_pers_no(int session_no)
  {
    array a = lookup(session_no);
    return a ? a[0]->person : -1;
  }

  array query_vars(int session_no, array(string) vars)
  {
    array a = lookup(session_no);
    WERROR(sprintf("Session_db->query_vars: a = %O\n", a));
    if (a)
      WERROR(sprintf("Session_db:query_vars: indices = %O\n",
		     "           vars = %O\n",
		     indices(a[0]), vars));

    return a && rows(a[0], vars);
  }
  
  void update(int n, mixed v)
  {
    cache::remember(n, v);
  }
#endif
  
  void store(int n, mixed v)
  {
    object key = cache::lock();
    cache::remember(n, v);
  }

  array(int) query_all_sessions()
  {
    return all_indices();
  }
}

class Globals
{
  inherit Thread.Mutex;
  object table;

  int conf_counter;
  int text_counter;

  array query_vars(array(string) vars)
  {
    object key = lock();
    return Array.map(vars, lambda(string s, object t)
			   { return t->get(s); },
		     table);
  }
  
  array set_var(string name, mixed value)
  {
    object key = lock();
    return table->set(name, value);
  }

  int current_text()
  {
    return text_counter;
  }
  
  int next_text()
  {
    object key = lock();
    text_counter++;
    table->set("text_counter", text_counter);
    return text_counter;
  }

  int current_conf()
  {
    return conf_counter;
  }
  
  int next_conf()
  {
    object key = lock();
    conf_counter++;
    table->set("conf_counter", conf_counter);
    return conf_counter;
  }

  void create(object db)
  {
    table = db->table("globals");
    text_counter = table->get("text_counter");
    conf_counter = table->get("conf_counter");
  }
}
    
void create()
{
  db = PDB.db(Config.query_var("location"), "rwcC");
  if (!db)
    throw( ({ "Database: Could not open database!\n", backtrace() }) );
  WERROR("Opened database.\n");
  texts = Text_db(db,Config.query_var("text_cache_size"));
  WERROR("Created Text_db.\n");
  text_stats = Text_Stat_db(db, Config.query_var("text_stat_cache_size"));
  WERROR("Created Text_Stat_db.\n");
  uconferences = UConf_db(db);
  WERROR("Created UConf_db.\n");  
  conferences = Conf_db(db);
  WERROR("Created Conf_db.\n");
  conf_texts = Conf_texts_db(db, Config.query_var("conf_texts_cache_size"));
  WERROR("Created Conf_texts_db.\n");  
  upersons = UPerson_db(db);
  WERROR("Created UPerson_db.\n");
  memberships = Memberships_db(db, Config.query_var("membership_cache_size"),
			       Config.query_var("membership_sync_time"));
  WERROR("Created Memberships_db.\n");
  created_texts
    = Person_texts_db(db, Config.query_var("created_texts_cache_size"));
  WERROR("Created Person_texts_db.\n");  
  unread = Person_unread_db(db, Config.query_var("unread_cache_size"),
			    Config.query_var("unread_sync_time"));
  WERROR("Created Person_unread_db.\n");
  marks = Mark_db(db, Config.query_var("mark_cache_size"));
  WERROR("Created Mark_db.\n");
  
  globals = Globals(db);
  
  sessions = Session_db();
  session_counter = Counter(0);

  // conf_counter = Persistent_Counter(globals, "conf_counter");
  // text_counter = Persistent_Counter(globals, "text_counter");
}

/* Diverse operationer dr flera objekt r inblandade. */

array create_conf(int creator, string name, int type)
{
  if (!Config.query_var("anyone_may_create_conf"))
    {
      array a = upersons->lookup(creator);
      if (!a || !(a[0]->privileges & Types.priv_create_conf))
	return ({ Errors.perm });
    }

  if (!strlen(name) || (strlen(name) > Config.query_var("max_name_length")))
    return ({ Errors.bad_name });

  int conf_no = globals->next_conf();

  /* Skapa uconf */
    object uconf = Server.UConference();
  uconf->name = name;
  uconf->type = type & ~Types.conf_letterbox;
  uconf->highest_local_no = 0;
  uconf->nice = Config.query_var("default_conference_nice");
  
  object conf = Server.Conference();
  conf->creation_time = time();
  conf->last_written = 0;
  conf->creator = creator;
  conf->supervisor = creator;
  conf->super_conf = creator;
  conf->presentation = 0;
  conf->permitted_submitters = (type & Types.conf_secret) && conf_no;
  conf->members = (< >);

  /* Skapa m|tet */
  if (!uconferences->create_conf(conf_no, uconf))
    return ({ Errors.conf_exists });

  conferences->store(conf_no, conf);

  upersons->record_created_conf(creator);

  return ({ 0, conf_no });
}

/* Returnerar ({ felkod }) eller ({ 0, pers_no }) */
array(int) create_person(int creator, string username,
		    string name, string crypted_passwd)
{
  if (!Config.query_var("create_person_without_login"))
    {
      if (!creator || !upersons->exists(creator))
	return ({ Errors.login });
    }
  
  if (!Config.query_var("anyone_may_create_person"))
    {
      /* Kolla privilegier */
      if (!creator)
	return ({ Errors.perm });
      
      array a = upersons->lookup(creator);
      if (!a || !(a[0]->privileges & Types.priv_create_pers))
	return ({ Errors.perm });
    }

  if (!strlen(name) || (strlen(name) > Config.query_var("max_name_length")))
    return ({ Errors.bad_name });
  
  int pers_no = globals->next_conf();

  /* Skapa brevlda */
  object uconf = Server.UConference();
  uconf->name = name;
  uconf->type = Types.conf_rd_prot | Types.conf_letterbox;
  uconf->highest_local_no = 0;
  uconf->nice = Config.query_var("default_letterbox_nice");
  
  object conf = Server.Conference();
  conf->creation_time = time();
  conf->last_written = 0;
  conf->creator = creator || pers_no;
  conf->supervisor = creator || pers_no;
  conf->super_conf = creator;
  conf->presentation = 0;
  conf->permitted_submitters = 0;
  conf->members = (< pers_no >);
  
  object membership = Server.Membership();
  membership->last_time_read = 0;
  membership->conference = pers_no;
  membership->priority = Config.query_var("default_priority");
  membership->first_unread = globals->current_text();
  membership->no_of_texts_read = 0;

  object m = Server->Person_memberships();
  m->memberships = ({ membership });

  object uperson = Server.UPerson();
  uperson->username = username;
  uperson->privileges = 0; /* Alla bittar r noll */
  uperson->flags = 0;
  uperson->last_login = 0;
  uperson->no_of_confs = 1;
  /* Alla vriga rknare startar p noll */

  uperson->passwd = crypted_passwd;

  if (!uconferences->create_conf(pers_no, uconf))
    return ({ Errors.conf_exists });

  conferences->store(pers_no, conf);
  memberships->store(pers_no, m);
  created_texts->store_new(pers_no);
  unread->store_new(pers_no);
  marks->store_new(pers_no);
  
  upersons->store(pers_no, uperson);

  if (creator)
    upersons->record_created_pers(creator);
  return ({ 0, pers_no });
}

int supervises(int conf, int person)
{
  int supervisor = conferences->query_supervisor(conf);
  return (supervisor > 0) && conferences->is_member(supervisor, person);
}

int may_access_conf(int conf, int person)
{
  if ( (conf == person) || ( (!uconferences->is_secret(conf))))
    return 1;
  if (!person)
    return 0;
  return supervises(conf, person);
}
  
int may_read_conf(int conf, int person)
{
  if ( (conf == person) || ( (!uconferences->is_protected(conf))))
    return 1;
  if (!person)
    return 0;

  if (conferences->is_member(conf, person))
    return 1;
  return supervises(conf, person);
}

int may_read_text(int text, int person)
{
  int author = text_stats->query_author(text);
  if (author < 0)
    return 0;
  if (supervises(author, person))
    return 1;
  
  array(int) recipients = text_stats->query_all_recipients(text);

  if (!recipients)
    return 0;

  foreach(recipients, int conf)
    {
      if (may_read_conf(conf, person))
	return 1;
    }
  return 0;
}

int add_member(int conf, int pers, int priority, int force, int where)
{
  array a = memberships->lookup(pers);
  if (!a)
    return Errors.undef_pers;

  int i;
  int size = sizeof(a[0]->memberships);
  if ((where < 0) || (where > size))
    {
      return Errors.index_out_of_range;
    }
  for(int i = 0; i<size; i++)
    if (a[0]->memberships[i]->conference == conf)
      break;
  if (i < size)
    { /* Redan medlem */
      if ((where < 0) || (where >= size))
	{
	  return Errors.index_out_of_range;
	}
      if (!force)
	return Errors.perm;
      if (where != i)
	{
	  object m =  a[0]->memberships[i];
	  a[0]->memberships = a[0]->memberships[..i-1]
                              + a[0]->memberships[i+1..];
	  a[0]->memberships = a[0]->memberships[..where-1]
		              + ({ m })
                              + a[0]->memberships[where..];
	}
      a[0]->memberships[where]->priority = priority;
      memberships->update(pers, a[0]);
    }
  else
    { /* Nytt medlemskap */
      if ((where < 0) || (where > size))
	{
	  return Errors.index_out_of_range;
	}
      object m = Server.Membership();
      m->last_time_read = 0;
      m->conference = conf;
      m->priority = priority;
      m->first_unread = conf_texts->query_first_text(conf);
      int no_of_texts_read = 0;

      a[0]->memberships = a[0]->memberships[..where-1]
	                  + ({ m })
                          + a[0]->memberships[where..];
      memberships->update(pers, a[0]);

      upersons->record_memberships(pers, sizeof(a[0]->memberships));
      conferences->add_member(conf, pers);
    }
  return 0;
}

int add_text_to_conf(int conf_no, int text_no)
{
  array a = conferences->lookup(conf_no);
  if (!a)
    return 0;

  int no_of_texts = conf_texts->add_text(conf_no, text_no);
  if (!no_of_texts)
    return 0;
  
  a[0]->no_of_texts = no_of_texts;
  int t = time();
  if (t < a[0]->last_written)
    werror("Database.add_text_to_conf: last_written in the future!\n");
  else
    a[0]->last_written = t;
  
  conferences->update(conf_no, a[0]);

  return 1;
}
  
